#define PROCESSING_TEXTURE_SHADER

#ifdef GL_ES
precision mediump float;
precision mediump int;
#endif

uniform sampler2D texture;
uniform vec2 texOffset;
uniform int type;

varying vec4 vertColor;
varying vec4 vertTexCoord;
mat3 matrix;


void main() {
	
	// matrix of all the colors to be used
	vec4 picMatrix[9];
	
	float iFloat = 0.0;
	// adding each color
	for(int i=0; i<9; i++) {
	
		float xAdjust; // the amount to adjust for each color
		float yAdjust;
		// change the adjust based on location
		// should end up with this config:
		// -, - | 0, - | +, -
		// -, 0 | 0, 0 | +, 0
		// -, + | 0, + | +, +
		if (mod(iFloat, 3.)==0.)
			xAdjust = -texOffset.s;
		else if (mod(iFloat, 3.)==1.)
			xAdjust = 0.;
		else 
			xAdjust = texOffset.s;
		
		if (i<3) 
			yAdjust = -texOffset.t;
		else if (i<6)
			yAdjust = 0.;
		else
			yAdjust = texOffset.t;
		
		// use the adjustments
		vec2 location = vertTexCoord.xy + vec2(xAdjust, yAdjust);
		// and load the color at that location from the texture
		picMatrix[i] = texture2D(texture, location);
		iFloat++;
	}
	// the sum of all the colors
	vec4 sum;
	
	// the boolean for if this is at an edge
	bool border = false;
	// if any color is black, this is at an edge
	for(int i = 0; i<9; i++) {
		
		bool xDiff = abs(picMatrix[i].x-picMatrix[4].x)>0.5;
		bool yDiff = abs(picMatrix[i].y-picMatrix[4].y)>0.5;
		bool zDiff = abs(picMatrix[i].z-picMatrix[4].z)>0.5;
		border = xDiff || yDiff || zDiff || border;
	}
	// if this is on a border, then draw it
	if (border)
		gl_FragColor = texture2D(texture, vertTexCoord.st);
	else {
		sum = vec4(0.0, 0.0, 0.0, 1.0);
		sum.w = 1.0;
		gl_FragColor = sum;
	}
}