import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.awt.event.KeyEvent; 
import java.util.LinkedList; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class CUBE extends PApplet {



Cube rubeCube; // the base cube
// the axis vectors for vector rotation
PVector kZ = new PVector(0, 0, 1);
PVector kY = new PVector(0, 1, 0);
PVector kX = new PVector(1, 0, 0);
int activeShader; // the reference to the current shader
ArrayList<PShader> shaders; // the shaders loaded into the program
PGraphics image, processedImage; // the images the shaders draw onto
boolean shaderOn = false; // is the shader active?
boolean hud = false; // is the heads-up display active?
boolean hudClicked = false; // was something on the hud clicked?
int cubesize = 50; // the default size of a cube
int cubespace = 10; // the default space between cubes
boolean goodGraphics = false; // are good graphics enabled?
int hudselection = 0; // the button selected on the hud
// the list of rotations to perform
LinkedList moves = new LinkedList<Integer>(); 
int waitframes = 0; // the number of frames to wait
final int waittime = 6; // number of frames to complete a rotation
boolean rotated = false; // has there been a move yet?
PFont rubeFont; // the loaded font
final int cwhite = color(255, 255, 255); // white
final int cyellow = color(255, 255, 0); // yellow
final int cblack = color(0, 0, 0); // black
final int cgreen = color(0, 192, 0); // green
final int cblue = color(0, 0, 255); // blue
final int cred = color(255, 0, 0); // red
final int corange = color(255, 150, 0); // orange
public void setup() {
  size(600, 600, P3D); // the size of this
  rubeFont = createFont("LiberationSans-Bold-48.vlw", 30.f); // loading the font
  // create a 3x3 cube array with locations at -1, 0, and 1 
  rubeCube = new Cube();
  for (int i = -1; i < 2; i++) {
    for (int j = -1; j < 2; j++) {
      for (int k = -1; k < 2; k++) {
        rubeCube.add(new SubCube(new PVector(i, j, k)));
      }
    }
  }
  // color the outside of the cube
  rubeCube.changeFaceColor("front", cyellow, 1, "z");
  rubeCube.changeFaceColor("back", cblack, -1, "z");
  rubeCube.changeFaceColor("left", cgreen, -1, "x");
  rubeCube.changeFaceColor("right", cblue, 1, "x");
  rubeCube.changeFaceColor("top", cred, -1, "y");
  rubeCube.changeFaceColor("bottom", corange, 1, "y");
  
  // load in some shaders
  shaders = new ArrayList<PShader>();
    shaders.add(loadShader("Night.glsl"));
    shaders.add(loadShader("Alert.glsl"));
    shaders.add(loadShader("Cyclone.glsl"));
    shaders.add(loadShader("DangerZone.glsl"));
    shaders.add(loadShader("Wave.glsl"));
    shaders.add(loadShader("Blind.glsl"));
  activeShader = 0; // start at the front of the list
  
  // create the images to draw to
  image = createGraphics(width, height, P3D);
  processedImage = createGraphics(width, height, P3D);
  noStroke();
  image.noStroke();
}

public void draw() {
  //rotate the cube if needed
  if (waitframes > 0)
    waitframes -= 1;
  else if (moves.size() > 0) {
    rubeCube.performRotation();
    waitframes = waittime;
  }
  // if good graphics are enabled, draw to the image. otherwise, 
  // draw to the screen.
  if (goodGraphics) {
    image.beginDraw(); 
    image.noStroke();
    image.textFont(rubeFont);
    beginCamera();
    camera(rubeCube.cam.loc.x, rubeCube.cam.loc.y, rubeCube.cam.loc.z, 
            width/2.0f, height/2.0f, 0, 0, 1, 0);
    endCamera();
    image.background(255); // white background
    image.pushMatrix();
    image.translate(width/2, height/2, 0); // center the matrix at the middle of the screen
    rubeCube.update(image); // update the Rube Cube
    // if the cube is in a win position and the cube has been moved, let the player know
    if(rubeCube.winCheckSum() && rotated) {
      image.fill(0);
      image.textSize(20);
      image.textAlign(CENTER);
      image.text("Solved", 0, 0, 150);
    }
    image.popMatrix();
    image.fill(0);
    image.textSize(32);
    image.textAlign(CENTER);
    image.text("THE RUBE CUBE", width/2, 50); // write the title
    image.endDraw();
    processedImage.beginDraw();
    // if the shader is on, process the image through the shader
    if(shaderOn) {
      processedImage.shader(shaders.get(activeShader));
      shaders.get(activeShader).set("time", millis()/1000.0f);
      if(activeShader == 2) { 
        shaders.get(2).set("midx", (float)width/2);
        shaders.get(2).set("midy", (float)height/2); 
      }  
    }
    else 
      processedImage.resetShader();
    processedImage.image(image, 0, 0);
    processedImage.endDraw();
    image(processedImage, 0, 0);
  } else {
    noStroke();
    textFont(rubeFont);
    beginCamera();
    camera(rubeCube.cam.loc.x, rubeCube.cam.loc.y, rubeCube.cam.loc.z,
            width/2.0f, height/2.0f, 0, 0, 1, 0);
    endCamera();
    background(255);
    pushMatrix();
    translate(width/2, height/2, 0);
    rubeCube.update();
    if(rubeCube.winCheckSum() && rotated) {
      fill(0);
      translate(0, 0, 150);
      box(100, 50, 0);
    }
    popMatrix();
    fill(0);
    textSize(32);
    textAlign(CENTER);
    text("THE RUBE CUBE", width/2, 50);
  }
  // if the heads-up display is on, display it.
  if (hud) {
    noStroke();
    fill(0, 128, 0);
    rect(100, height-51, 150, 2);
    rect(width-250, height-51, 150, 2);
    fill(128, 0, 0);
    rect(95 + cubesize * 3 / 2, height-60, 10, 20);
    rect(width-255 + cubespace * 3 / 2, height-60, 10, 20);
    fill(128, 128, 192);
    rect(25, 300, 100, 50);
    rect(25, 200, 100, 50);
    rect(25, 400, 100, 50);
    rect(25, 475, 25, 25);
    rect(100, 475, 25, 25);
    rect(width-125, 200, 100, 50);
    rect(width-125, 300, 100, 50);
    rect(width-125, 400, 100, 50);
    fill(255);
    triangle(45, 480, 45, 496, 30, 488);
    triangle(105, 480, 105, 496, 120, 488);
    if (true) {
      fill(0, 0, 128);
      textSize(16);
      textAlign(CENTER);
      text("Cube Size", 175, height-15);
      text("Cube Spacing", width-175, height-15);
      textSize(24);
      fill(255, 255, 255);
      text("Shuffle", 75, 235);
      text("Solve", 75, 335);
      text("Shader", 75, 435);
      text("Checker", width-75, 235);
      text("Flower", width-75, 335);
      text("Middle", width-75, 435);
      textSize(24);
    } else {
      fill(0, 0, 128);
      rect(125, height-30, 100, 20);
      rect(width-225, height-30, 100, 20);
      fill(255, 255, 255);
      rect(35, 210, 80, 30);
      rect(35, 310, 80, 30);
      rect(35, 410, 80, 30);
      rect(width-115, 210, 80, 30);
      rect(width-115, 310, 80, 30);
      rect(width-115, 410, 80, 30);
    }
  }
  // show the options button always
  fill(128, 128, 192);
  rect(25, 25, 100, 50);
  fill(255);
  textSize(24);
  textAlign(CENTER);
  if (true)
    text("Options", 75, 60);
  else
    rect(45, 37, 60, 25);
}


// if the mouse is dragged and the hud is up, update that. Otherwise rotate the cube.
public void mouseDragged() {
  if (hud)
    updateHUD();
  if (!hudClicked)
    rubeCube.rotateCube();
}

// handles mouse clicks
public void mousePressed() {
  // if the options button is clicked, toggle the hud
  if (mouseX >= 25 && mouseX < 125 && mouseY >= 25 && mouseY < 75) {
    hudselection = 11;
    updateHUD();
  }
  // if the hud is up, check if any buttons were clicked
  else if (hud) {
    if (mouseX >= 100 && mouseX < 250 && mouseY >= height-60 && mouseY < height-40)
      hudselection = 1;
    else if (mouseX >= width-250 && mouseX < width-100 && mouseY >= height-60 && mouseY < height-40)
      hudselection = 2;
    else if (mouseX >= 25 && mouseX < 125 && mouseY >= 200 && mouseY < 250)
      hudselection = 3;
    else if (mouseX >= 25 && mouseX < 125 && mouseY >= 300 && mouseY < 350)
      hudselection = 4;
    else if (mouseX >= width-125 && mouseX < width-25 && mouseY >= 200 && mouseY < 250)
      hudselection = 5;
    else if (mouseX >= width-125 && mouseX < width-25 && mouseY >= 300 && mouseY < 350)
      hudselection = 6;
    else if (mouseX >= width-125 && mouseX < width-25 && mouseY >= 400 && mouseY < 450)
      hudselection = 7;
    else if (mouseX >= 25 && mouseX <= 125 && mouseY >= 400 && mouseY <= 450)
      hudselection = 8;
    else if (mouseX >= 25 && mouseX <= 50 && mouseY >= 475 && mouseY <= 500) 
      hudselection = 9;
    else if (mouseX >= 100 && mouseX <= 125 && mouseY >= 475 && mouseY <= 500)
      hudselection = 10;
    else {
      // if nothing on the hud was clicked, update the cube and mouse
      rubeCube.mouse.mouseClick();
      rubeCube.mouseClick(); 
    }
    updateHUD();
  } else {
    // if the hud is not up, have the rubecube and the mouse act on the click
    rubeCube.mouse.mouseClick();
    rubeCube.mouseClick();
  }
}

// handles when the mouse is released
public void mouseReleased() {
    // release the mouse from a hud click
    hudselection = 0;
    updateHUD();
    // update the cube
    rubeCube.mouseRelease();
}

// handles key strokes
public void keyPressed() {
  // if s and goodgraphics are on, toggle the shaders
  if(key == 's') {
    goodGraphics = true;
    shaderOn =! shaderOn; 
    //println(shaderOn);
  } 
  if(key == CODED) {
    // if up is pressed, change the shader reference by +1
    if(keyCode == UP) {
      activeShader = (activeShader + 1) % shaders.size();
    }
    // if down is pressed, change the shader reference by -1
    if(keyCode == DOWN) {
      if(activeShader == 0)
        activeShader = shaders.size() - 1;
      else
        activeShader--;
    }
  }
  // if shift, toggle the hud
  if (keyCode == SHIFT) {
    hud = !hud;
  }
  // if backspace, set the cube size to default
  if (keyCode == BACKSPACE) {
    cubesize = 50;
    cubespace = 10;
  }
  // if g, toggle good graphics
  if (keyCode == KeyEvent.VK_G) {
    goodGraphics = !goodGraphics; 
  }
  // if W, check for a win and print to the command line the result
  if (keyCode == KeyEvent.VK_W) {
    boolean check = rubeCube.winCheckSum();
    if(check)
      println("SOLVED!!");
    else
      println("Not solved...  Keep trying!");
  }
}

// rotates about an axis based on Rodrigues' rotation formula
public PVector rotateAboutAxis(PVector k, PVector v, float theta) {
   
   return PVector.add(PVector.add(PVector.mult(v, cos(theta)), 
                                  PVector.mult(k.cross(v), sin(theta))), 
                      PVector.mult(k, k.dot(v) * (1 - cos(theta))));
}

// rotate about all the given axes by the given angles
public PVector rotateAboutAxes(PVector k, PVector j, PVector l, PVector v,
                      PVector thetas) {
  
  return rotateAboutAxis(k, rotateAboutAxis(j, 
                            rotateAboutAxis(l, v, thetas.z), -thetas.y), 
                            thetas.x);
  
}

// checks for a collision of the given location on the plane of the given triangle
public boolean triangleCollision(PVector colLoc, PVector a, PVector b, PVector c) {
    PVector colA = PVector.sub(colLoc, a);
    PVector bA = PVector.sub(b, a);
    PVector cA = PVector.sub(c, a);
    
    float coldotb = colA.dot(bA);
    float coldotc = colA.dot(cA);
    float bdotc = bA.dot(cA);
    float bdotb = bA.dot(bA);
    float cdotc = cA.dot(cA);
    
    float denom = bdotb * cdotc - bdotc * bdotc;
    float alpha = (bdotb * coldotc - bdotc * coldotb) / denom;
    float beta = (cdotc * coldotb - bdotc * coldotc) / denom;
    return alpha >= 0 && beta >= 0 && alpha + beta < 1;
}

// gives if the first set is contained in the second
public boolean setContains(ArrayList u, ArrayList v) {
  if(u.size()==0) {
    return true;
  } else if (v.size()==0) {
    return false;
  } else {
    boolean equal = v.contains(u.get(0));
    u.remove(0);
    return equal 
        && setContains(u, v);
  }
}

//get the positive x mod y
public float posMod(float x, float y) {
  return (x % y + y) % y;
}

//update stuff from the HUD
public void updateHUD() {
  if (hudselection == 0) {
    // the hud was not clicked
    hudClicked = false;
  } else if (hudselection == 1) {
    // change the cube size
    cubesize = Math.max(2, Math.min(100, (mouseX - 100) / 3 * 2));
    hudClicked = true;
    //println("cubesize = " + cubesize + ", cubespace = " + cubespace);
  } else if (hudselection == 2) {
    // change the space between cubes
    cubespace = Math.max(2, Math.min(100, (mouseX - width+250) / 3 * 2));
    hudClicked = true;
    //println("cubesize = " + cubesize + ", cubespace = " + cubespace);
  } else if (hudselection == 3 && !hudClicked) {
    // randomize the cube
    //can't edit the cube while it's in the middle of being edited
    if (moves.size() <= 0 && waitframes <= 0) {
      //add a random number of random shuffles
      for (int count = (int)(Math.random() * 20) + 15; count > 0; count -= 1) {
        addRandomMove();
      }
    }
    hudClicked = true;
    //println("Rotating " + shuffles + " times");
  } else if (hudselection == 4) {
    // solve the cube
    //can't edit the cube while it's in the middle of being edited
    if (moves.size() <= 0 && waitframes <= 0 && !hudClicked)
      //println("Solving the cube");
      solve();
    hudClicked=true;
  } else if (hudselection == 5  && !hudClicked) {
    // create a checkboard pattern
    checkerBoardPattern();
    hudClicked=true; 
  } else if (hudselection == 6  && !hudClicked) {
    // create a flower pattern
    checkerFlowerPattern();
    hudClicked=true; 
  } else if (hudselection == 7  && !hudClicked) {
    // create a switched middle pattern
    middleSwitchPattern();
    hudClicked=true; 
  } else if (hudselection == 8  && !hudClicked) {
    // turn the shader on or off
    goodGraphics = true;
    shaderOn = !shaderOn;
    hudClicked=true; 
  } else if (hudselection == 9  && !hudClicked) {
    // go to the last shader
    if (activeShader == 0) activeShader = shaders.size()-1;
    else activeShader--;
    hudClicked=true;
  } else if (hudselection == 10  && !hudClicked) {
    // go to the next shader
    activeShader = (activeShader+1) % shaders.size();
    hudClicked=true;
  } else if (hudselection == 11 && !hudClicked) {
    // toggle the hud on or off
    hud=!hud;
    hudClicked=true;
  }
}

//get a random twist of the rubik's cube
public void addRandomMove() {
  addRotation((int)(Math.random() * 3) - 1, (int)(Math.random() * 3), 
                (int)(Math.random() * 2) * 2 + 1);
}

//add a rotation to the rotation queue
public int addRotation(int row, int direction, int times) {
  int value = (row + 1) + (direction << 2) + (times << 4);
  moves.add(value);
  return value;
}

class Cube {
 ArrayList<SubCube> cubes; // the cubes in this
 float rotY, rotX, rotZ; // the rotation in space of this
 mouseHandler mouse; // the mouse handler
 Camera cam; // the camera pointed at the cube
 String activeFace; // the selected face
 PVector activeCube; // the selected cube
 boolean grabbedCube; // is the cube grabbed by the mouse? 
 
 Cube() {
   cubes = new ArrayList<SubCube>(); // there are no cubes when created
   mouse = new mouseHandler(); // default mouse handler
   cam = new Camera();
   // this has no rotation at first
   rotY = 0;
   rotX = 0;
   rotZ = 0;
   activeFace = "None"; // there is no active face
   activeCube = new PVector(0, 0, 0); // placeholder for the active cube
   grabbedCube = false; // the cube is not grabbed
 }
 
 // add cubes to the array in this
 public void add(SubCube subCube) {
   cubes.add(subCube);
 }
 
 // update this on the given image
 public boolean update(PGraphics image) {
   // update the mouse
   mouse.updateMouse();
   
   // rotate the matrix
   image.pushMatrix();
   image.rotateX(rotX);
   image.rotateY(-rotY);
   image.rotateZ(rotZ);
   
   // update the cubes
   for(SubCube sc : cubes) {
     sc.update(image);
   }
   
   image.popMatrix();
   return false;
 }
 
 // update this
 public boolean update() {
   // update the mouse information
   mouse.updateMouse();
   
   // save the current matrix, then rotate this
   pushMatrix();
   rotateX(rotX);
   rotateY(-rotY);
   rotateZ(rotZ);
   
   //update the subcubes
   for(SubCube sc : cubes) {
     sc.update();
   }
   
   popMatrix();
   return false;
 }
  
  // checks for a win based on the check sum
  public boolean winCheckSum() {
    // initialize the sums
    float checkFront = 0.f;
    float checkBack = 0.f;
    float checkLeft = 0.f;
    float checkRight = 0.f;
    float checkTop = 0.f;
    float checkBottom = 0.f;
    
    // add the cube values to the sums
    for (SubCube sc : cubes) {
      checkFront += sc.cFront;
      checkBack += sc.cBack;
      checkLeft += sc.cLeft;
      checkRight += sc.cRight;
      checkTop += sc.cTop;
      checkBottom += sc.cBottom;
    }
    // add the checks and sums to their respective arrays
    ArrayList<Float> sums = new ArrayList<Float>();
    ArrayList<Float> checks = new ArrayList<Float>();
    sums.add(checkFront);
    checks.add(27.f);
    sums.add(checkBack);
    checks.add(63.f);
    sums.add(checkTop);
    checks.add(117.f);
    sums.add(checkBottom);
    checks.add(153.f);
    sums.add(checkLeft);
    checks.add(207.f);
    sums.add(checkRight);
    checks.add(261.f);
    // return if all the sums are in the checks
    return setContains(sums, checks);
  }
  
  // checks if there is a win config based on the colors
  public boolean winCheckColor() {
    int ctop = cwhite;
    int cbottom = cwhite;
    int cleft = cwhite;
    int cright = cwhite;
    int cfront = cwhite;
    int cback = cwhite;
    for (SubCube sc: cubes) {
      //if the current color is white, replace it
      if (ctop == cwhite)
        ctop = sc.top;
      //if the cube color is different from this color and neither are white, 
      //it's not solved
      else if (sc.top != cwhite && sc.top != ctop)
        return false;
      if (cbottom == cwhite)
        cbottom = sc.bottom;
      else if (sc.bottom != cwhite && sc.bottom != cbottom)
        return false;
      if (cleft == cwhite)
        cleft = sc.left;
      else if (sc.left != cwhite && sc.left != cleft)
        return false;
      if (cright == cwhite)
        cright = sc.right;
      else if (sc.right != cwhite && sc.right != cright)
        return false;
      if (cfront == cwhite)
        cfront = sc.front;
      else if (sc.front != cwhite && sc.front != cfront)
        return false;
      if (cback == cwhite)
        cback = sc.back;
      else if (sc.back != cwhite && sc.back != cback)
        return false;
    }
    return true;
  }
  
  //changes the face if it matches the location on the given axis
  public void changeFaceColor(String face, int change, int loc, String axis) {
    for(SubCube sc : cubes) {
      if(axis.equals("z") && sc.currentPos.z == loc
      || axis.equals("y") && sc.currentPos.y == loc
      || axis.equals("x") && sc.currentPos.x == loc) {
        sc.changeFace(face, change);
      }
    }
  }
  
  //changes the face of a specific cube based on the active face
  public void changeActiveFace(int change, PVector loc) {
    for(SubCube sc : cubes) {  
      if(sc.currentPos.z == loc.z && sc.currentPos.y == loc.y
            && sc.currentPos.x == loc.x) {
          sc.changeFace(activeFace, change);
      }
      
    }
  }
  
  // changes the color of every face in this
  public void changeFaceColor(String face, int change) {
     for(SubCube sc : cubes) {
         sc.changeFace(face, change);
     }
  }
  
  // rotate the given row about the Z axis
  public void rotateRowZ(int loc) { 
    performRotation(addRotation(loc, 2, 1));
  }
  
  // rotate the given row about the Z axis in the opposite directions
  public void rotateReverseRowZ(int loc) { 
    performRotation(addRotation(loc, 2, 3));
  }
  
  // rotate the given row about the X axis
  public void rotateRowX(int loc) { 
    performRotation(addRotation(loc, 0, 1));
  }
  
  // rotate the given row about the X axis in the opposite directions
  public void rotateReverseRowX(int loc) { 
    performRotation(addRotation(loc, 0, 3));
  }
  
  // rotate the given row about the Y axis
  public void rotateRowY(int loc) {
    performRotation(addRotation(loc, 1, 1));
  }
  
  // rotate the given row about the Y axis in the opposite directions
  public void rotateReverseRowY(int loc) { 
    performRotation(addRotation(loc, 1, 3));
  }
  
  // the user rotates the given row about the Z axis
  public void userRotateRowZ(int loc) { 
    addRotation(loc, 2, 1);
  }
  
  // the user rotates the given row about the Z axis in the opposite directions
  public void userRotateReverseRowZ(int loc) { 
    addRotation(loc, 2, 3);
  }
  
  // the user rotates the given row about the X axis
  public void userRotateRowX(int loc) { 
    addRotation(loc, 0, 1);
  }
  
  // the user rotates the given row about the X axis in the opposite directions
  public void userRotateReverseRowX(int loc) { 
    addRotation(loc, 0, 3);
  }
  
  // the user rotates the given row about the Y axis
  public void userRotateRowY(int loc) {
    addRotation(loc, 1, 1);
  }
  
  // the user rotates the given row about the Y axis in the opposite directions
  public void userRotateReverseRowY(int loc) { 
    addRotation(loc, 1, 3);
  }
  
  // if the mouse is clicked, see if it clicked on a cube
  public void mouseClick() {
    cam.rayTrace(mouse.xLoc, mouse.yLoc, this);
    // if there is no active face, the cube was not grabbed
    grabbedCube =! activeFace.equals("None");
  }
  
  // handle mouse releases
  public void mouseRelease() {
    // the cube is no longer grabbed regardless
    grabbedCube = false;
    // have the mouse do its thing on a release
    rubeCube.mouse.mouseRelease();
  }
  
  // rotate either the whole cube or a row
  public void rotateCube() {
    // if the cube is not grabbed, rotate the cube in space
   if(!grabbedCube) {  
     rotX = posMod(rotX + (mouse.yLoc - mouseY) / 500, 2 * PI);
     //println("" + rotX);
     if (rotX < PI / 2 || rotX >= 3 * PI / 2)
       rotY = posMod(rotY + (mouse.xLoc - mouseX) / 500, 2 * PI);
     else
       rotY = posMod(rotY + (mouseX - mouse.xLoc) / 500, 2 * PI);
   }
   // if it is grabbed and the swipe vector is greater than the cubesize, rotate a row
   else if(sqrt(sq(mouse.xDist)+sq(mouse.yDist))>cubesize) {
     PVector swipe = new PVector(mouse.xDist, mouse.yDist, 0);
     swipe.normalize();
     rotateRow(swipe);
   } 
  }
  
  // rotates a row of the cube
  public void rotateRow(PVector swipe) {
    //can't edit the cube while it's in the middle of being edited
    if (moves.size() > 0 || waitframes > 0)
      return;
    // generate the relevant corners of the cube
    PVector cornerFTL = rotateAboutAxes(kX, kY, kZ, 
                     new PVector(-cubesize / 2, cubesize / 2, cubesize / 2), 
                     new PVector(rotX, rotY, rotZ));
    PVector cornerFTR = rotateAboutAxes(kX, kY, kZ, 
                     new PVector(cubesize / 2, cubesize / 2, cubesize / 2), 
                     new PVector(rotX, rotY, rotZ));
    PVector cornerFBL = rotateAboutAxes(kX, kY, kZ,
                     new PVector(-cubesize / 2, -cubesize / 2, cubesize / 2), 
                     new PVector(rotX, rotY, rotZ));
    PVector cornerBTL = rotateAboutAxes(kX, kY, kZ, 
                     new PVector(-cubesize / 2, cubesize / 2, -cubesize / 2), 
                     new PVector(rotX, rotY, rotZ));
    // generate the axes of the cube based on the corners
    PVector xDirection = PVector.sub(cornerFTR, cornerFTL);
    PVector yDirection = PVector.sub(cornerFBL, cornerFTL);
    PVector zDirection = PVector.sub(cornerBTL, cornerFTL);
    // remove the z and then normalize
    xDirection.z = 0;
    yDirection.z = 0;
    zDirection.z = 0;
    xDirection.normalize();
    yDirection.normalize();
    zDirection.normalize();
    // generate the angles between the vectors and the swipes
    float angleX = acos(xDirection.dot(swipe));
    float angleY = acos(yDirection.dot(swipe));
    float angleZ = acos(zDirection.dot(swipe));
    /* 
      Rotate the cube based on the selected face:
        if front or back, compare to x and y
        if left or right, compare to z and y
        if top or bottom, compare to x and z
      If the angle to a direction is less than PI/4, then
      rotate in that direction.
      After the rotations are done, set the active face to "none"
    */
    if(activeFace.equals("front")) {
      if(angleX <= PI/4   && angleX<=angleY) {
        userRotateRowY((int)activeCube.y);
      } else if(angleY <= PI/4) {
        userRotateRowX((int)activeCube.x);
      } else if(angleX >= PI * 3/4   && angleX>=angleY) {
        userRotateReverseRowY((int)activeCube.y);
      } else if(angleY >= PI * 3/4) {
        userRotateReverseRowX((int)activeCube.x);
      }
      activeFace = "None";
    } else if(activeFace.equals("back")) {
      if(angleX <= PI/4   && angleX<=angleY) {
        userRotateReverseRowY((int)activeCube.y);
      } else if(angleY <= PI/4) {
        userRotateReverseRowX((int)activeCube.x);
      } else if(angleX >= PI * 3/4   && angleX>=angleY) {
        userRotateRowY((int)activeCube.y);
      } else if(angleY >= PI * 3/4) {
        userRotateRowX((int)activeCube.x);
      }
      activeFace = "None";
    } else if(activeFace.equals("top")) {
      if(angleX <= PI/4 && angleX<=angleZ) {
        userRotateRowZ((int)activeCube.z);
      } else if(angleZ <= PI/4) {
        userRotateRowX((int)activeCube.x);
      } else if(angleX >= PI * 3/4  && angleX>=angleZ) {
        userRotateReverseRowZ((int)activeCube.z);
      } else if(angleZ >= PI * 3/4) {
        userRotateReverseRowX((int)activeCube.x);
      }
      activeFace = "None";
    } else if(activeFace.equals("bottom")) {
      if(angleX <= PI/4  && angleX<=angleZ) {
        userRotateReverseRowZ((int)activeCube.z);
      } else if(angleZ <= PI/4) {
        userRotateReverseRowX((int)activeCube.x);
      } else if(angleX >= PI * 3/4  && angleX>=angleZ) {
        userRotateRowZ((int)activeCube.z);
      } else if(angleZ >= PI * 3/4) {
        userRotateRowX((int)activeCube.x);
      }
      activeFace = "None";
    } else if(activeFace.equals("left")) {
      if(angleZ <= PI/4  && angleZ<=angleY) {
        userRotateReverseRowY((int)activeCube.y);
      } else if(angleY <= PI/4) {
        userRotateRowZ((int)activeCube.z);
      } else if(angleZ >= PI * 3/4  && angleZ>=angleY) {
        userRotateRowY((int)activeCube.y);
      } else if(angleY >= PI * 3/4) {
        userRotateReverseRowZ((int)activeCube.z);
      }
      activeFace = "None";
    } else if(activeFace.equals("right")) {
      if(angleZ <= PI/4  && angleZ<=angleY) {
        userRotateRowY((int)activeCube.y);
      } else if(angleY <= PI/4) {
        userRotateReverseRowZ((int)activeCube.z);
      } else if(angleZ >= PI * 3/4  && angleZ>=angleY) {
        userRotateReverseRowY((int)activeCube.y);
      } else if(angleY >= PI * 3/4) {
        userRotateRowZ((int)activeCube.z);
      }
      activeFace = "None";
    }
  }
  
  // return a list of the cubes in this
  public String toString() {
    return cubes.toString();
  }
  
  //return a clone of this cube that can be used to solve
  public Cube solveableClone() {
    Cube c = new Cube();
    for (SubCube sc : cubes) {
      c.cubes.add(sc.clone());
    }
    return c;
  }
  
  //rotate the cube in the specified way
  public void performRotation() {
    performRotation((Integer)(moves.remove()));
  }
  
  //rotate the cube in the specified way
  public void performRotation(int rotation) {
    //0xTTDDRR
    //TT == times to rotate
    //DD == direction to rotate
    //RR == row to rotate( + 1)
    int row = (rotation & 3) - 1;
    int direction = (rotation >> 2 & 3);
    int times = (rotation >> 4 & 3);
    for (int t = 0; t < times; t += 1) {
      //x
      if (direction == 0) {
        for(SubCube sc : cubes) {
          if(sc.currentPos.x == row) {
            float curY = sc.currentPos.y;
            float curZ = sc.currentPos.z;
            sc.currentPos.z = curY;
            sc.currentPos.y = -curZ;
            sc.moveColorsX();
            sc.rotX -= waittime;
            if (sc.rotX <= -waittime * 2)
              sc.rotX += waittime * 4;
          }
        }
      //y
      } else if (direction == 1) {
        for(SubCube sc : cubes) {
          if(sc.currentPos.y == row) {
            float curZ = sc.currentPos.z;
            float curX = sc.currentPos.x;
            sc.currentPos.x = curZ;
            sc.currentPos.z = -curX;
            sc.moveColorsY();
            sc.rotY -= waittime;
            if (sc.rotY <= -waittime * 2)
              sc.rotY += waittime * 4;
          }
        }
      //z
      } else {
        for(SubCube sc : cubes) {
          if(sc.currentPos.z == row) {
            float curX = sc.currentPos.x;
            float curY = sc.currentPos.y;
            sc.currentPos.x = -curY;
            sc.currentPos.y = curX;
            sc.moveColorsZ();
            sc.rotZ -= waittime;
            if (sc.rotZ <= -waittime * 2)
              sc.rotZ += waittime * 4;
          }
        }
      }
    }
    rotated = true;
  }
}

// the component cubes
class SubCube {
  int left, right, front, back, top, bottom; // the colors of the faces
  float cLeft, cRight, cFront, cBack, cTop, cBottom; // the values of the faces
  int rotX, rotY, rotZ; // the rotation of this
  PVector currentPos; // the current location of this
  
  SubCube(PVector currentPos) {
    // the color of each side is white by default
    left = cwhite;
    right = cwhite;
    front = cwhite;
    back = cwhite;
    top = cwhite;
    bottom = cwhite;
    // the rotation is 0 by default
    rotX = 0;
    rotY = 0;
    rotZ = 0;
    this.currentPos=currentPos;
    // checksum of all the locations
    cLeft = max(-currentPos.x*3, 0);
    cRight = max(currentPos.x*7, 0);
    cFront = max(currentPos.z*13, 0);
    cBack = max(-currentPos.z*17, 0);
    cTop = max(-currentPos.y*23, 0);
    cBottom = max(currentPos.y*29, 0);
  }
  
  // list the colors of the faces
  public int[] makeListOfFaces() {
    int[] list = new int[6];
    list[0] = front;
    list[1] = back;
    list[2] = left;
    list[3] = right;
    list[4] = top;
    list[5] = bottom;
    return list;
  }
  
  // updates this
  public boolean update() {
    drawBox();
    return false;
  }
  
  // updates this on the given image
  public boolean update(PGraphics image) {
    drawBox(image);
    return false;
  }
  
  // draw this on the given image
  public void drawBox(PGraphics image) {
    image.pushMatrix();
    // rotate this by its rotation, then update the rotation
    image.rotateX(PI * rotX / waittime / 2);
    image.rotateY(PI * rotY / waittime / 2);
    image.rotateZ(PI * rotZ / waittime / 2);
    if (rotX > 0)
      rotX -= 1;
    else if (rotX < 0)
      rotX += 1;
    if (rotY > 0)
      rotY -= 1;
    else if (rotY < 0)
      rotY += 1;
    if (rotZ > 0)
      rotZ -= 1;
    else if (rotZ < 0)
      rotZ += 1;
      // translate this to its location
      image.translate((cubesize + cubespace) * currentPos.x, 
                      (cubesize + cubespace) * currentPos.y, 
                      (cubesize + cubespace) * currentPos.z);
      // then draw each face of the cube.
      // each face has to be drawn seperately so they can be different colors
      image.translate(0, 0, cubesize / 2);
      image.fill(front);
      image.box(cubesize, cubesize, 0);
      image.translate(0, 0, -cubesize);
      image.fill(back);
      image.box(cubesize, cubesize, 0);
      image.translate(-cubesize / 2, 0, cubesize / 2);
      image.fill(left);
      image.box(0, cubesize, cubesize);
      image.translate(cubesize, 0, 0);
      image.fill(right);
      image.box(0, cubesize, cubesize);
      image.translate(-cubesize / 2, -cubesize / 2, 0);
      image.fill(top);
      image.box(cubesize, 0, cubesize);
      image.translate(0, cubesize, 0);
      image.fill(bottom);
      image.box(cubesize, 0, cubesize);
      image.popMatrix();
  }
  
  // draw this onto the screen
  public void drawBox() {
      pushMatrix();
    rotateX(PI * rotX / waittime / 2);
    rotateY(PI * rotY / waittime / 2);
    rotateZ(PI * rotZ / waittime / 2);
    if (rotX > 0)
      rotX -= 1;
    else if (rotX < 0)
      rotX += 1;
    if (rotY > 0)
      rotY -= 1;
    else if (rotY < 0)
      rotY += 1;
    if (rotZ > 0)
      rotZ -= 1;
    else if (rotZ < 0)
      rotZ += 1;
      translate((cubesize + cubespace) * currentPos.x, 
                (cubesize + cubespace) * currentPos.y, 
                (cubesize + cubespace) * currentPos.z);
      translate(0, 0, cubesize / 2);
      fill(front);
      box(cubesize, cubesize, 0);
      translate(0, 0, -cubesize);
      fill(back);
      box(cubesize, cubesize, 0);
      translate(-cubesize / 2, 0, cubesize / 2);
      fill(left);
      box(0, cubesize, cubesize);
      translate(cubesize, 0, 0);
      fill(right);
      box(0, cubesize, cubesize);
      translate(-cubesize / 2, -cubesize / 2, 0);
      fill(top);
      box(cubesize, 0, cubesize);
      translate(0, cubesize, 0);
      fill(bottom);
      box(cubesize, 0, cubesize);
      popMatrix();
  }
  
  // move the colors about the X axis
  public void moveColorsX() { 
    int curTop = top;
    int curFront = front;
    int curBottom = bottom;
    int curBack = back;
    float curTopCor = cTop;
    float curBottomCor = cBottom;
    float curFrontCor = cFront;
    float curBackCor = cBack;
    
    top = curFront;
    front = curBottom;
    bottom = curBack;
    back = curTop;
    cTop = curFrontCor;
    cFront = curBottomCor;
    cBottom = curBackCor;
    cBack = curTopCor;
  }
  
  // move the colors about the Y axis
  public void moveColorsY() { 
    int curLeft = left;
    int curFront = front;
    int curRight = right;
    int curBack = back;
    float curLeftCor = cLeft;
    float curRightCor = cRight;
    float curFrontCor = cFront;
    float curBackCor = cBack;
    
    left = curBack;
    front = curLeft;
    right = curFront;
    back = curRight;
    cLeft = curBackCor;
    cFront = curLeftCor;
    cRight = curFrontCor;
    cBack = curRightCor;
  }
  
  // move the colors along the Z axis
  public void moveColorsZ() {
    int curTop = top;
    int curLeft = left;
    int curBottom = bottom;
    int curRight = right;
    float curTopCor = cTop;
    float curBottomCor = cBottom;
    float curLeftCor = cLeft;
    float curRightCor = cRight;
    
    top = curLeft;
    right = curTop;
    bottom = curRight;
    left = curBottom;
    cTop = curLeftCor;
    cRight = curTopCor;
    cBottom = curRightCor;
    cLeft = curBottomCor;
  }
  
  // change the given face to the given color
  public void changeFace(String face, int change) {
    
    if(face.equals("left")) {
      left = change;
    }
    else if(face.equals("right")) {
      right = change;
    } 
    else if(face.equals("top")) {
      top = change;
    } 
    else if(face.equals("bottom")) {
      bottom = change;
    } 
    else if(face.equals("front")) {
      front = change;
    } 
    else if(face.equals("back")) {
      back = change;
    } 
  }
  
  // checks if this has collided with the given ray
  public void collide(PVector loc, PVector ray, PVector rot,  
                        ArrayList<PVector> collideLoc, 
                        ArrayList<PVector> collideCube, 
                        ArrayList<String> collideFace) {
    // generate the cube's location and its 8 corners
    PVector cubeLocation = new PVector(currentPos.x * (cubesize + cubespace), 
                        currentPos.y * -(cubesize + cubespace), 
                        currentPos.z * (cubesize + cubespace));
    PVector cornerFTR = rotateAboutAxes(kX, kY, kZ, 
                        new PVector(cubeLocation.x + cubesize / 2, 
                        cubeLocation.y + cubesize / 2, 
                        cubeLocation.z + cubesize / 2), rot);
    cornerFTR.add(new PVector(width/2, height/2, 0));
    PVector cornerFTL = rotateAboutAxes(kX, kY, kZ, 
                        new PVector(cubeLocation.x - cubesize / 2, 
                        cubeLocation.y + cubesize / 2, 
                        cubeLocation.z + cubesize / 2), rot);
    cornerFTL.add(new PVector(width/2, height/2, 0));
    PVector cornerFBR = rotateAboutAxes(kX, kY, kZ, 
                        new PVector(cubeLocation.x + cubesize / 2, 
                        cubeLocation.y - cubesize / 2,
                        cubeLocation.z + cubesize / 2), rot);
    cornerFBR.add(new PVector(width/2, height/2, 0));
    PVector cornerFBL = rotateAboutAxes(kX, kY, kZ, 
                        new PVector(cubeLocation.x - cubesize / 2, 
                        cubeLocation.y - cubesize / 2, 
                        cubeLocation.z + cubesize / 2), rot);
    cornerFBL.add(new PVector(width/2, height/2, 0));
    PVector cornerBTR = rotateAboutAxes(kX, kY, kZ,  
                        new PVector(cubeLocation.x + cubesize / 2, 
                        cubeLocation.y + cubesize / 2, 
                        cubeLocation.z - cubesize / 2), rot);
    cornerBTR.add(new PVector(width/2, height/2, 0));
    PVector cornerBTL = rotateAboutAxes(kX, kY, kZ,  
                        new PVector(cubeLocation.x - cubesize / 2,
                        cubeLocation.y + cubesize / 2, 
                        cubeLocation.z - cubesize / 2), rot);
    cornerBTL.add(new PVector(width/2, height/2, 0));
    PVector cornerBBR = rotateAboutAxes(kX, kY, kZ,  
                        new PVector(cubeLocation.x + cubesize / 2, 
                        cubeLocation.y - cubesize / 2, 
                        cubeLocation.z - cubesize / 2), rot);
    cornerBBR.add(new PVector(width/2, height/2, 0));
    PVector cornerBBL = rotateAboutAxes(kX, kY, kZ,  
                        new PVector(cubeLocation.x - cubesize / 2, 
                        cubeLocation.y - cubesize / 2, 
                        cubeLocation.z - cubesize / 2), rot);
    cornerBBL.add(new PVector(width/2, height/2, 0));
    // generate the normals based on the corner's locations
    PVector frontNormal =  PVector.sub(cornerFTR, cornerBTR);
    PVector backNormal =   PVector.sub(cornerBTR, cornerFTR);
    PVector leftNormal =   PVector.sub(cornerFTL, cornerFTR);
    PVector rightNormal =  PVector.sub(cornerFTR, cornerFTL);
    PVector bottomNormal = PVector.sub(cornerFBR, cornerFTR);
    PVector topNormal =    PVector.sub(cornerFTR, cornerFBR);
    
    // if the ray collided with the front, add this and the front to the collision arrays
    if(collideFace(frontNormal, cornerFTL, cornerFTR, cornerFBL, 
                    cornerFBR, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("front");
    }
    
    // do the same for each other face
    if(collideFace(backNormal, cornerBTR, cornerBTL, cornerBBR, 
                    cornerBBL, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("back");
    }
    
    if(collideFace(topNormal, cornerBTL, cornerBTR, cornerFTL, 
                    cornerFTR, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("top");
    }
    
    if(collideFace(bottomNormal, cornerFBL, cornerFBR, cornerBBL, 
                    cornerBBR, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("bottom");
    }
    
    if(collideFace(rightNormal, cornerBTR, cornerFTR, cornerBBR, 
                    cornerFBR, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("right");
    }
    
    if(collideFace(leftNormal, cornerFTL, cornerBTL, cornerFBL,
                    cornerBBL, loc, ray, collideLoc)) {
      collideCube.add(this.currentPos);
      collideFace.add("left");
    }
  }
  
  // detect if a ray hit this face based on the given normal and corners
  public boolean collideFace(PVector normal, PVector tl, PVector tr, 
                      PVector bl, PVector br, PVector camLoc, PVector ray, 
                      ArrayList<PVector> collideLoc) {
    // generate numbers for collision detection
    float d = normal.dot(tl);
    float normalDotRay = normal.dot(ray);
    
    // if parallel the ray did not hit
    if (normalDotRay == 0) 
      return false;
    
    // if it did intersect then do triangle collisions on the two triangles that cover this cube 
    float intersect = (d - camLoc.dot(normal)) / normalDotRay;
    PVector colLoc = PVector.add(camLoc, PVector.mult(ray, intersect));
    if(triangleCollision(colLoc, tl, bl, tr)
        || triangleCollision(colLoc, br, tr, bl)) {
      collideLoc.add(PVector.add(colLoc, PVector.mult(ray, intersect)));
      return true;
    }
    return false;
  }
  
  // creates a string based on this
  public String toString() {
    return currentPos.toString();
  }
  
  // a clone of this subcube
  public SubCube clone() {
    SubCube s = new SubCube(new PVector(currentPos.x, currentPos.y, 
                              currentPos.z));
    s.left = left;
    s.right = right;
    s.front = front;
    s.back = back;
    s.top = top;
    s.bottom = bottom;
    s.rotX = rotX;
    s.rotY = rotY;
    s.rotZ = rotZ;
    return s;
  }
}

// the camera class
class Camera {
  PVector w; // forward vector
  PVector u; // down vector
  PVector v; // left vector
  PVector loc; // camera location
  PVector left; // viewfield left vector 
  PVector down; // viewfield down vector
  float d = 10.0f; // distance to the screen
  float tanTheta = tan(PI/3); // field of view angle
  PVector up = new PVector(0, 1, 0); // the up vector
  
  Camera() {
   // generate the vectors
   loc = new PVector(width/2, height/2, (height/2.0f) / tan(PI*30.0f / 180.0f));
   w = new PVector(0, 0, -1);
   w.normalize();
   u = w.cross(up);
   u.normalize();
   v = w.cross(u);
   v.normalize();
   left = PVector.sub(PVector.mult(w, d), PVector.mult(u, d * tanTheta));
   left.add(loc);
   down = PVector.sub(PVector.mult(w, d), PVector.mult(v, d * tanTheta));
   down.add(loc); 
  }
  
  // casts a ray at the given cube for any collisions
  public void rayTrace(float x, float y, Cube cube) {
    PVector ray = new PVector(0, 0, 0); // ray to be cast
    PVector distX = new PVector(0, 0, 0); // the X and Y components of the cast ray 
    PVector distY = new PVector(0, 0, 0);
    PVector.add(left, PVector.mult(u, 2*d*tanTheta*(x+0.5f)/width), distX);
    PVector.add(down, PVector.mult(v, 2*d*tanTheta*(y+0.5f)/height), distY);
    distX.sub(loc);
    distY.sub(loc);
    PVector.add(PVector.add(distX, distY), PVector.mult(w, d), ray);
    ray.normalize();
    PVector rot = new PVector(-cube.rotX, cube.rotY, -cube.rotZ);
    
    // collision locations, cubes, and faces
    ArrayList<PVector> collideLoc = new ArrayList<PVector>();
    ArrayList<PVector> collideCube = new ArrayList<PVector>();
    ArrayList<String> collideFace = new ArrayList<String>();
    // collide with each cube
    for(SubCube sc : cube.cubes) {
       sc.collide(loc, ray, rot, collideLoc, collideCube, collideFace);
    }
    // determine the closest collision
    if(collideLoc.size()>0) {
      PVector currentClosest = collideLoc.get(0);
      float closestMagnitude = PVector.sub(currentClosest, loc).mag();
      PVector currentCube = collideCube.get(0);
      String currentFace = collideFace.get(0);
      int i=0;
      for(PVector l : collideLoc) {
       float magnitude = PVector.sub(l, loc).mag();
       if(magnitude >= d
       && magnitude < closestMagnitude
       && PVector.sub(currentClosest, loc).dot(w) > 0) {
         currentClosest = l;
         currentCube = collideCube.get(i);
         currentFace = collideFace.get(i);
         closestMagnitude = magnitude;
       }
       i++;
      }
      if(PVector.sub(currentClosest, loc).mag() < d
      || PVector.sub(currentClosest, loc).dot(w) < 0) {
        // if there were no collisions in range, there is no active face
        cube.activeFace = "None";
        return;
      }
      // if there was, make those collisions active
      cube.activeFace = currentFace;
      cube.activeCube = currentCube;
      return;
    }
    // if there were no collisions, mark the active face as none
    cube.activeFace = "None";
  }
  
  // raytracer that debugs hitboxes based on the face it hits. 
  // entirely the same as above but it returns a color instead of nothing
  public int hitboxDebug(float x, float y, Cube cube) {
    PVector ray = new PVector(0, 0, 0);
    PVector distX = new PVector(0, 0, 0);
    PVector distY = new PVector(0, 0, 0);
    PVector.add(left, PVector.mult(u, 2*d*tanTheta*(x+0.5f)/width), distX);
    PVector.add(down, PVector.mult(v, 2*d*tanTheta*(y+0.5f)/height), distY);
    distX.sub(loc);
    distY.sub(loc);
    PVector.add(PVector.add(distX, distY), PVector.mult(w, d), ray);
    ray.normalize();
    PVector rot = new PVector(-cube.rotX, cube.rotY, -cube.rotZ);
    
    ArrayList<PVector> collideLoc = new ArrayList<PVector>();
    ArrayList<PVector> collideCube = new ArrayList<PVector>();
    ArrayList<String> collideFace = new ArrayList<String>();
    for(SubCube sc : cube.cubes) {
       sc.collide(loc, ray, rot, collideLoc, collideCube, collideFace);
    }
    
    if(collideLoc.size() > 0) {
      PVector currentClosest = collideLoc.get(0);
      float closestMagnitude = PVector.sub(currentClosest, loc).mag();
      PVector currentCube = collideCube.get(0);
      String currentFace = collideFace.get(0);
      int i = 0;
      for(PVector l : collideLoc) {
       float magnitude = PVector.sub(l, loc).mag();
       if(magnitude >= d && magnitude < closestMagnitude
       /*&& PVector.sub(currentClosest, loc).dot(w)<0*/) {
         currentClosest = l;
         currentCube = collideCube.get(i);
         currentFace = collideFace.get(i);
         closestMagnitude = magnitude;
       }
       i++;
      }
      if(PVector.sub(currentClosest, loc).mag() < d
      || PVector.sub(currentClosest, loc).dot(w) < 0) {
        return color(255);
      }
      if(currentFace.equals("front"))
        return color(255, 0, 0);
      else if (currentFace.equals("back"))
        return color(255, 255, 0);
      else if (currentFace.equals("top"))
        return color(0, 255, 0);
      else if (currentFace.equals("bottom"))
        return color(0, 255, 255);
      else if (currentFace.equals("left"))
        return color(0, 0, 255);
      else if (currentFace.equals("right"))
        return color(255, 0, 255);
      else
        return color(255);
    }
    cube.activeFace = "None";
    return color(255);
  }
}

// class for handling mouse movements
class mouseHandler {
 float xLoc, yLoc, xClick, yClick, xDist, yDist; // location, click location, and the distance betweeen the two
 boolean mouseDown; // is the mouse down?

  // build the mouse information based off the current mouse location
  mouseHandler() {
    xLoc = mouseX;
    yLoc = mouseY;
    xClick = 0;
    yClick = 0;
    xDist = 0;
    yDist = 0;
    mouseDown = false;
  }
  
  // updates the mouse information in this
  public void updateMouse() {
    xLoc = mouseX;
    yLoc = mouseY;
    if (mouseDown) {
      xDist = xLoc - xClick;
      yDist = yLoc - yClick;
    }
  }
  
  // if the mouse was clicked change the relevant information
  public void mouseClick() {
    xClick = mouseX;
    yClick = mouseY;
    mouseDown = true;
  }
  
  // if the mouse was released change the relevant information
  public void mouseRelease() {
    xClick = 0;
    yClick = 0;
    xDist = 0;
    yDist = 0;
    mouseDown = false;
  }
}

/* 
  The AI begins here.  Each method is appropriately named in a way 
  that comments are rendered redundant.  Each move is a specific 
  sequence of steps required to make the move without disturbing 
  previously placed pieces.  The AI begins by solving the bottom 
  face and the bottom row at the same time, then solves the middle 
  row, and then finishes with the top face and top row.
*/

public void solve() {
  Cube oldcube = rubeCube;
  rubeCube = rubeCube.solveableClone();
  solveBottom();
  solveSecondRow();
  solveTop(); 
  rubeCube = oldcube;
}

public void solveBottom() {
  if(!bottomSolved()) {
    firstBottomCorner();
    secondBottomCorner();
    thirdBottomCorner();
    fourthBottomCorner();
    firstBottomEdge();
    secondBottomEdge();
    thirdBottomEdge();
    fourthBottomEdge();
    while(!bottomCenterSolved()) {
      solveCenter();
    }
  }
}

public void fourthBottomEdge() {
  SubCube cube = findFourthBottomEdge();
  PVector pos = cube.currentPos;
  while(!pos.equals(new PVector(1, 1, 0))) {
    if(pos.y != 0) {
      while(pos.y == 1) {
        rubeCube.rotateReverseRowZ(0);
        rubeCube.rotateRowY(-1);
        rubeCube.rotateRowZ(0);
      }
      while(!pos.equals(new PVector(0, -1, -1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowZ(0);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(0);
    }
    else {
      while(!pos.equals(new PVector(-1, 0, -1))) {
       rubeCube.rotateRowY(0);
      } 
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowY(0);
      rubeCube.rotateRowY(0);
      rubeCube.rotateReverseRowX(1);
      
    }
  }
  
  while(cube.bottom != corange) {
    rubeCube.rotateReverseRowZ(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(0);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(0);
  }

}

public SubCube findFourthBottomEdge() {
  int orange = 0;
  int blue = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    blue = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cblue)
        blue += 1; 
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(orange == 1 && blue == 1 && white == 4)
      return sc;
  } 
  return null;
}

public void thirdBottomEdge() {
  SubCube cube = findThirdBottomEdge();
  PVector pos = cube.currentPos;
  while(!pos.equals(new PVector(0, 1, 1))) {
    if(pos.y != 0) {
      while(pos.y == 1) {
         if(pos.x == 1) {
           rubeCube.rotateReverseRowZ(0);
           rubeCube.rotateRowY(-1);
           rubeCube.rotateRowZ(0);
         }
         else {
           rubeCube.rotateRowX(0);
           rubeCube.rotateRowY(-1);
           rubeCube.rotateReverseRowX(0);
         }
      }
      while(!pos.equals(new PVector(1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowX(0);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(0);
    }
    else {
      while(!pos.equals(new PVector(1, 0, -1))) {
       rubeCube.rotateRowY(0);
      } 
      rubeCube.rotateRowZ(1);
      rubeCube.rotateReverseRowY(0);
      rubeCube.rotateReverseRowZ(1);
      
    }
  }
  
  while(cube.bottom != corange) {
    rubeCube.rotateRowX(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(0);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(0);
  }

}

public SubCube findThirdBottomEdge() {
  int orange = 0;
  int yellow = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    yellow = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cyellow)
        yellow += 1; 
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(orange == 1 && yellow == 1 && white == 4)
      return sc;
  } 
  return null;
}

public void secondBottomEdge() {
  SubCube cube = findSecondBottomEdge();
  PVector pos = cube.currentPos;
  while(!pos.equals(new PVector(-1, 1, 0))) {
    if(pos.y != 0) {
      while(pos.y == 1) {
         if(pos.x == 0) {
           rubeCube.rotateRowX(0);
           rubeCube.rotateRowY(-1);
           rubeCube.rotateReverseRowX(0);
         }
         else {
           rubeCube.rotateReverseRowZ(0);
           rubeCube.rotateRowY(-1);
           rubeCube.rotateRowZ(0);
         }
      }
      while(!pos.equals(new PVector(0, -1, 1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowZ(0);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(0);
    }
    else {
      while(!pos.equals(new PVector(1, 0, 1))) {
       rubeCube.rotateRowY(0);
      } 
      rubeCube.rotateRowX(-1);
      rubeCube.rotateReverseRowY(0);
      rubeCube.rotateReverseRowX(-1);
      
    }
  }
  
  while(cube.bottom != corange) {
    rubeCube.rotateRowZ(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(0);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(0);
  }

}

public SubCube findSecondBottomEdge() {
  int orange = 0;
  int green = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    green = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cgreen)
        green += 1; 
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(orange == 1 && green == 1 && white == 4)
      return sc;
  } 
  return null;
}

public void firstBottomEdge() {
  SubCube cube = findFirstBottomEdge();
  PVector pos = cube.currentPos;
  while(!pos.equals(new PVector(0, 1, -1))) {
    if(pos.y != 0) {
      while(pos.y == 1) {
         if(pos.x == 0) {
           rubeCube.rotateRowX(0);
         }
         else if(pos.z == 0 && pos.x == -1) {
           rubeCube.rotateRowZ(0);
         }
         else {
           rubeCube.rotateReverseRowZ(0);
         }
      }
      while(!pos.equals(new PVector(-1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowX(0);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(0);
    }
    else {
      while(!pos.equals(new PVector(-1, 0, 1))) {
       rubeCube.rotateRowY(0);
      } 
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateReverseRowY(0);
      rubeCube.rotateReverseRowZ(-1);
    }
  }
  while(cube.bottom != corange) {
    rubeCube.rotateReverseRowX(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(0);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(0);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(0);
  }
}

public SubCube findFirstBottomEdge() {
  int orange = 0;
  int black = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    black = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cblack)
        black += 1; 
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(orange == 1 && black == 1 && white == 4)
      return sc;
  } 
  return null;
}

public void fourthBottomCorner() {
  SubCube cube = findFourthBottomCorner();
  PVector pos = cube.currentPos;
  // place the corner
  while(!pos.equals(new PVector(1, 1, -1))) {
    while(!pos.equals(new PVector(1, -1, -1))) {
      rubeCube.rotateRowY(-1); 
    }
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
  }
    // orient the corner
    
  while(cube.bottom != corange) {
    rubeCube.rotateReverseRowZ(-1); 
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(-1); 
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(-1); 
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(-1);
  }
}

public SubCube findFourthBottomCorner() {
  int orange = 0;
  int black = 0;
  int blue = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    black = 0;
    blue = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cblack)
        black += 1; 
      else if(colors[i] == cblue)
        blue += 1; 
    }
    if(orange == 1 && black == 1 && blue == 1)
      return sc;
  } 
  return null;
}

public void thirdBottomCorner() {
  SubCube cube = findThirdBottomCorner();
  PVector pos = cube.currentPos;
  // place the corner
  while(!pos.equals(new PVector(1, 1, 1))) {
    while(pos.y == 1) {
      rubeCube.rotateRowX(1);
    }
    while(!pos.equals(new PVector(1, -1, 1))) {
      rubeCube.rotateRowY(-1); 
    }
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
  }
    // orient the corner
    
  while(cube.bottom != corange) {
    rubeCube.rotateRowX(1); 
    rubeCube.rotateRowX(1);    
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(1);
  }
}

public SubCube findThirdBottomCorner() {
  int orange = 0;
  int yellow = 0;
  int blue = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    yellow = 0;
    blue = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cyellow)
        yellow += 1; 
      else if(colors[i] == cblue)
        blue += 1; 
    }
    if(orange == 1 && yellow == 1 && blue == 1)
      return sc;
  } 
  return null;
}

public void secondBottomCorner() {
  SubCube cube = findSecondBottomCorner();
  PVector pos = cube.currentPos;
  // place the corner
  while(!pos.equals(new PVector(-1, 1, 1))) {
    while(pos.x == 1 && pos.y == 1) {
      rubeCube.rotateRowX(1);
    }
    while(!pos.equals(new PVector(-1, -1, 1))) {
      rubeCube.rotateRowY(-1); 
    }
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(1);
  }
    // orient the corner
    
  while(cube.bottom != corange) {
    rubeCube.rotateRowZ(1);     
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(1);
    rubeCube.rotateRowZ(1);
  }
}

public SubCube findSecondBottomCorner() {
  int orange = 0;
  int yellow = 0;
  int green = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    yellow = 0;
    green = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cyellow)
        yellow += 1; 
      else if(colors[i] == cgreen)
        green += 1; 
    }
    if(orange == 1 && yellow == 1 && green == 1)
      return sc;
  } 
  return null;
}

public void firstBottomCorner() {
  SubCube cube = findFirstBottomCorner();
  PVector first = cube.currentPos;
  // place the corner
  while(!first.equals(new PVector(-1, 1, -1))) {
    while(first.x != -1) {
      rubeCube.rotateRowZ((int)first.z); 
    }
    while(first.y != 1) {
      rubeCube.rotateRowZ((int)first.z); 
    }
    while(first.z != -1) {
      rubeCube.rotateRowY((int)first.z); 
    }
    cube = findFirstBottomCorner();
    first = cube.currentPos;
  }
  
  // orient the corner
  while(cube.bottom != corange) {
    rubeCube.rotateRowX(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    rubeCube.rotateRowX(-1);
  }
}

public SubCube findFirstBottomCorner() {
  int orange = 0;
  int black = 0;
  int green = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    black = 0;
    green = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cblack)
        black += 1; 
      else if(colors[i] == cgreen)
        green += 1; 
    }
    if(orange == 1 && black == 1 && green == 1)
      return sc;
  } 
  return null;
}

public void solveCenter() {
  PVector orangePos = findOrangeCenter();
  while(!orangePos.equals(new PVector(0, 1, 0))) {
    if(orangePos.equals(new PVector(0, -1, 0))) {
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowX(0);
      rubeCube.rotateRowX(0);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateReverseRowZ(1);
    }
    else if(orangePos.x > 0) {
      rubeCube.rotateRowX(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowZ(0);
      rubeCube.rotateReverseRowX(-1);
      rubeCube.rotateReverseRowX(1);
    }
    else if(orangePos.x < 0) {
      rubeCube.rotateRowX(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowZ(0);
      rubeCube.rotateReverseRowX(-1);
      rubeCube.rotateReverseRowX(1);
    }
    else if(orangePos.z > 0) {
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowX(0);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateReverseRowZ(1);
    }
    else if(orangePos.z < 0) {
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateReverseRowX(0);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateReverseRowZ(1);
    }
    
    orangePos = findOrangeCenter();
  }
}

public PVector findOrangeCenter() {
  int orange = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    orange = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == corange)
        orange += 1;
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(orange == 1 && white == 5)
      return sc.currentPos;
  } 
  return null;
}

public boolean bottomCenterSolved() {
  for(SubCube sc: rubeCube.cubes) {
    if(sc.currentPos.y == 1 && sc.currentPos.x == 0 && sc.currentPos.z == 0) {
      if(sc.bottom != corange) {
        return false;
      }
    }
  } 
  return true;
}

public boolean bottomSolved() {
  for(SubCube sc : rubeCube.cubes) {
     if(sc.currentPos.y == 1) {
        if(sc.bottom != corange)
          return false;
     }  
  }
  return true;
}

public void solveSecondRow() {
  solveMiddles();
  firstMiddleEdge();
  secondMiddleEdge();
  thirdMiddleEdge();
  fourthMiddleEdge();
}

public void fourthMiddleEdge() {
  SubCube cube = findFourthMiddleEdge();
  PVector pos = cube.currentPos;
  float x = pos.x;
  float y = pos.y;
  float z = pos.z;
  
  while(!pos.equals(new PVector(1, 0, -1))) {
    
    // move into position
    if(cube.top == cblack) {
      while(!pos.equals(new PVector(1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
    }
    else if(cube.top == cblue) {
      while(!pos.equals(new PVector(0, -1, -1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowZ(-1);
    }
  }
  if(cube.right == cblack && cube.back == cblue) {
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
    while(!pos.equals(new PVector(0, -1, -1))) {
      rubeCube.rotateRowY(-1);
    }
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
  }
}

public SubCube findFourthMiddleEdge() {
  int black = 0;
  int blue = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    black = 0;
    blue = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cblack)
        black += 1;
      else if(colors[i] == cwhite)
        white += 1; 
      else if(colors[i] == cblue)
        blue += 1; 
    }
    if(black == 1 && white == 4 && blue == 1)
      return sc;
  } 
  return null;
}

public void thirdMiddleEdge() {
  SubCube cube = findThirdMiddleEdge();
  PVector pos = cube.currentPos;
  float x = pos.x;
  float y = pos.y;
  float z = pos.z;
  
  while(!pos.equals(new PVector(1, 0, 1))) {
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    
    // move into position
    if(cube.top == cyellow) {
      while(!pos.equals(new PVector(1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
    }
    else if(cube.top == cblue) {
      while(!pos.equals(new PVector(0, -1, 1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
    }
  }
  if(cube.front == cblue && cube.right == cyellow) {
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(1);
    while(!pos.equals(new PVector(0, -1, 1))) {
      rubeCube.rotateRowY(-1);
    }
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(1);
  }
}

public SubCube findThirdMiddleEdge() {
  int yellow = 0;
  int blue = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    yellow = 0;
    blue = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cyellow)
        yellow += 1;
      else if(colors[i] == cwhite)
        white += 1; 
      else if(colors[i] == cblue)
        blue += 1; 
    }
    if(yellow == 1 && white == 4 && blue == 1)
      return sc;
  } 
  return null;
}

public void secondMiddleEdge() {
  SubCube cube = findSecondMiddleEdge();
  PVector pos = cube.currentPos;
  float x = pos.x;
  float y = pos.y;
  float z = pos.z;
  
  while(!pos.equals(new PVector(-1, 0, 1))) {
    if(y == 0 && -x == z) {
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
    }
    else if(y == 0 && x == z) {
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
    }
    
    // move into position
    if(cube.top == cgreen) {
      while(!pos.equals(new PVector(0, -1, 1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowX(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
    }
    else if(cube.top == cyellow) {
      while(!pos.equals(new PVector(-1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowX(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(-1);
    }
  }
  if(cube.front == cgreen && cube.left == cyellow) {
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(-1);
    while(!pos.equals(new PVector(-1, -1, 0))) {
      rubeCube.rotateRowY(-1);
    }
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowZ(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowZ(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(-1);
  }
}

public SubCube findSecondMiddleEdge() {
  int yellow = 0;
  int green = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    yellow = 0;
    green = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cyellow)
        yellow += 1;
      else if(colors[i] == cwhite)
        white += 1; 
      else if(colors[i] == cgreen)
        green += 1; 
    }
    if(yellow == 1 && white == 4 && green == 1)
      return sc;
  } 
  return null;
}

public void firstMiddleEdge() {
  SubCube cube = findFirstMiddleEdge();
  PVector pos = cube.currentPos;
  float x = pos.x;
  float y = pos.y;
  float z = pos.z;
  
  while(!pos.equals(new PVector(-1, 0, -1))) {
   if(y == 0 && -x == z) {
      rubeCube.rotateReverseRowY(-1);
      if(z == -1) {
        rubeCube.rotateReverseRowZ(-1);
      }
      else {
        rubeCube.rotateRowZ(1);
      }
      rubeCube.rotateRowY(-1);
      if(z == -1) {
        rubeCube.rotateRowZ(-1);
      }
      else {
        rubeCube.rotateReverseRowZ(1);
      }
      rubeCube.rotateRowY(-1);
      if(x == -1) {
        rubeCube.rotateRowX(-1);
      }
      else {
        rubeCube.rotateReverseRowX(1);
      }
      rubeCube.rotateReverseRowY(-1);
      if(x == -1) {
        rubeCube.rotateReverseRowX(-1);
      }
      else {
        rubeCube.rotateRowX(1);
      }
    }
    else if(y == 0 && x == z) {
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowZ(1);
    }
    
    // move into position
    if(cube.top == cgreen) {
      while(!pos.equals(new PVector(0, -1, -1))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowX(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(-1);
    }
    else if(cube.top == cblack) {
      while(!pos.equals(new PVector(-1, -1, 0))) {
        rubeCube.rotateRowY(-1);
      }
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowZ(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowZ(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(-1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateRowX(-1);
    }
  }
  if(cube.back == cgreen && cube.left == cblack) {
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    while(!pos.equals(new PVector(-1, -1, 0))) {
      rubeCube.rotateRowY(-1);
    }
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowZ(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowZ(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(-1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
  }
}

public SubCube findFirstMiddleEdge() {
  int black = 0;
  int green = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    black = 0;
    green = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cblack)
        black += 1;
      else if(colors[i] == cwhite)
        white += 1; 
      else if(colors[i] == cgreen)
        green += 1; 
    }
    if(black == 1 && white == 4 && green == 1)
      return sc;
  } 
  return null;
}

public void solveMiddles() {
  SubCube cube = findBlueCenter();
  PVector pos = cube.currentPos;
  while(!pos.equals(new PVector(1, 0, 0))) {
    rubeCube.rotateRowY(0);
  }
}

public SubCube findBlueCenter() {
  int blue = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    blue = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cblue)
        blue += 1;
      else if(colors[i] == cwhite)
        white += 1; 
    }
    if(blue == 1 && white == 5)
      return sc;
  } 
  return null;
}

public void solveTop() {
  solveRedTops();
  positionRedEdges();
  positionTopCorners();
  flipTopCorners();
}

public void flipTopCorners() {
   ArrayList<SubCube> cubes = getRedCorners();
   for(int i = 0; i < cubes.size(); i++) {
     SubCube cube = cubes.get(i);
     PVector pos = cube.currentPos;
     while(cube.top != cred) {
       while(!pos.equals(new PVector(1, -1, 1))) {
         rubeCube.rotateRowY(-1); 
       }
       rubeCube.rotateReverseRowX(1);
       rubeCube.rotateRowY(1);
       rubeCube.rotateRowX(1);
       rubeCube.rotateReverseRowY(1);
       rubeCube.rotateReverseRowX(1);
       rubeCube.rotateRowY(1);
       rubeCube.rotateRowX(1);
       rubeCube.rotateReverseRowY(1);
     }
   }
   repositionFrontEdge();
}

public ArrayList<SubCube> getRedCorners() {
  ArrayList<SubCube> list = new ArrayList<SubCube>();
  for(SubCube sc: rubeCube.cubes) {
    PVector pos = sc.currentPos;
    if(pos.equals(new PVector(1, -1, 1)))
      list.add(sc);
    else if(pos.equals(new PVector(-1, -1, -1)))
      list.add(sc);
    else if(pos.equals(new PVector(-1, -1, 1)))
      list.add(sc);
    else if(pos.equals(new PVector(1, -1, -1)))
      list.add(sc);
  }
  return list;
}

public void positionTopCorners() {
  SubCube cube = findCorrectCorner();
  while(cube == null) {
    repositionFrontEdge();
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(-1);
    cube = findCorrectCorner();
  }
  repositionFrontEdge();
  PVector pos = cube.currentPos;
  while(!cornersSolved()) {
    while(!cube.currentPos.equals(new PVector(1, -1, 1))) {
      rubeCube.rotateRowY(-1); 
    }
    
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateRowX(-1);
    rubeCube.rotateReverseRowY(-1);
    rubeCube.rotateReverseRowX(1);
    rubeCube.rotateRowY(-1);
    rubeCube.rotateReverseRowX(-1);
    while(!cube.currentPos.equals(pos)) {
      rubeCube.rotateRowY(-1); 
    }
    repositionFrontEdge();
  }
}

public boolean cornersSolved() {
  SubCube blackBlue = findRedBlackBlueCorner();
  SubCube blackGreen = findRedBlackGreenCorner();
  SubCube greenYellow = findRedGreenYellowCorner();
  SubCube yellowBlue = findRedYellowBlueCorner();
  return blackBlue.currentPos.equals(new PVector(1, -1, -1)) &&
          blackGreen.currentPos.equals(new PVector(-1, -1, -1)) &&
          greenYellow.currentPos.equals(new PVector(-1, -1, 1)) &&
          yellowBlue.currentPos.equals(new PVector(1, -1, 1));
}

public SubCube findRedBlackBlueCorner() {
  int red = 0;
  int blue = 0;
  int black = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    int[] colors = sc.makeListOfFaces();
    red = 0;
    blue = 0;
    black = 0;
    white = 0;
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cred) 
        red++;
      else if(colors[i] == cblack) 
        black++;
      else if(colors[i] == cblue)
        blue++;
      else if(colors[i] == cwhite)
        white++;
    }
    if(red == 1 && black == 1 && blue == 1 & white == 3)
      return sc;
  } 
  return null;
}

public SubCube findRedBlackGreenCorner() {
  int red = 0;
  int green = 0;
  int black = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    int[] colors = sc.makeListOfFaces();
    red = 0;
    green = 0;
    black = 0;
    white = 0;
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cred) 
        red++;
      else if(colors[i] == cblack) 
        black++;
      else if(colors[i] == cgreen)
        green++;
      else if(colors[i] == cwhite)
        white++;
    }
    if(red == 1 && black == 1 && green == 1 & white == 3)
      return sc;
  } 
  return null;
}

public SubCube findRedGreenYellowCorner() {
  int red = 0;
  int green = 0;
  int yellow = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    int[] colors = sc.makeListOfFaces();
    red = 0;
    green = 0;
    yellow = 0;
    white = 0;
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cred) 
        red++;
      else if(colors[i] == cyellow) 
        yellow++;
      else if(colors[i] == cgreen)
        green++;
      else if(colors[i] == cwhite)
        white++;
    }
    if(red == 1 && yellow == 1 && green == 1 & white == 3)
      return sc;
  } 
  return null;
}

public SubCube findRedYellowBlueCorner() {
  int red = 0;
  int blue = 0;
  int yellow = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    int[] colors = sc.makeListOfFaces();
    red = 0;
    blue = 0;
    yellow = 0;
    white = 0;
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cred) 
        red++;
      else if(colors[i] == cyellow) 
        yellow++;
      else if(colors[i] == cblue)
        blue++;
      else if(colors[i] == cwhite)
        white++;
    }
    if(red == 1 && yellow == 1 && blue == 1 & white == 3)
      return sc;
  } 
  return null;
}

public SubCube findCorrectCorner() {
  SubCube blackBlue = findRedBlackBlueCorner();
  SubCube blackGreen = findRedBlackGreenCorner();
  SubCube greenYellow = findRedGreenYellowCorner();
  SubCube yellowBlue = findRedYellowBlueCorner();
  if(blackBlue.currentPos.equals(new PVector(1, -1, -1))) {
    return blackBlue;
  }
  else if(blackGreen.currentPos.equals(new PVector(-1, -1, -1))) {
    return blackGreen;
  }
  else if(greenYellow.currentPos.equals(new PVector(-1, -1, 1))) {
    return greenYellow;
  }
  else if(yellowBlue.currentPos.equals(new PVector(1, -1, 1))) {
    return yellowBlue;
  }
  else
    return null;
  
}

public void positionRedEdges() {
  repositionFrontEdge();
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  while(right.right != cblue || left.left != cgreen
       || back.back != cblack || front.front != cyellow) {
    if(isPair()) {
      solvePair(findPair());
      repositionFrontEdge();
    }
    else if(oppositePair()) {
      solveOppositePair(findOppositePair());
      repositionFrontEdge();
    }
    else {
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(1);
      repositionFrontEdge();
    }
    front = getFrontTopEdge();
    back = getBackTopEdge();
    left = getLeftTopEdge();
    right = getRightTopEdge();
  }
   
}

public void solvePair(int[] pair) {
  SubCube left = getLeftTopEdge();
  while(left.left != pair[0]) {
    rubeCube.rotateRowY(-1); 
    left = getLeftTopEdge();
  }
  rubeCube.rotateRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowX(1);
}

public void solveOppositePair(int[] pair) {
  SubCube right = getRightTopEdge();
  while(right.right != pair[0]) {
    rubeCube.rotateRowY(-1); 
    right = getRightTopEdge();
  }
  rubeCube.rotateRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateRowX(1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowY(-1);
  rubeCube.rotateReverseRowX(1);
}

public int[] findOppositePair() {
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  if(front.front == cyellow && left.left != cgreen
      && right.right != cblue && back.back == cblack)
    return new int[]{front.front, back.back};
  else if(front.front == cblack && left.left != cgreen
        && right.right != cblue && back.back == cyellow)
    return new int[]{front.front, back.back};
  else if(front.front != cyellow && left.left == cgreen
          && right.right == cblue && back.back != cblack)
    return new int[]{left.left, right.right};
  else if(front.front != cyellow && left.left == cblue
          && right.right == cgreen && back.back != cblack)
    return new int[]{left.left, right.right};
  else
    return null;
}

public int[] findPair() {
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  if(front.front == cyellow && left.left == cgreen
      && right.right != cblue && back.back != cblack)
    return new int[]{cgreen, cyellow};
  else if(front.front == cyellow && left.left != cgreen
          && right.right == cblue && back.back != cblack)
    return new int[]{cyellow, cblue};
  else if(front.front != cyellow && left.left != cgreen
          && right.right == cblue && back.back == cblack)
    return new int[]{cblue, cblack};
  else if(front.front != cyellow && left.left == cgreen
          && right.right != cblue && back.back == cblack)
    return new int[]{cblack, cgreen};
  else
    return null;
}

public boolean isPair() {
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  if(front.front == cyellow && left.left == cgreen
      && right.right != cblue && back.back != cblack)
    return true;
  else if(front.front == cyellow && left.left != cgreen
          && right.right == cblue && back.back != cblack)
    return true;
  else if(front.front != cyellow && left.left != cgreen
          && right.right == cblue && back.back == cblack)
    return true;
  else if(front.front != cyellow && left.left == cgreen
          && right.right != cblue && back.back == cblack)
    return true;
  else
    return false;
}

public boolean oppositePair() {
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  if(front.front == cyellow && left.left != cgreen
      && right.right != cblue && back.back == cblack)
    return true;
  else if(front.front == cblack && left.left != cgreen
        && right.right != cblue && back.back == cyellow)
    return true;
  else if(front.front != cyellow && left.left == cgreen
          && right.right == cblue && back.back != cblack)
    return true;
  else if(front.front != cyellow && left.left == cblue
          && right.right == cgreen && back.back != cblack)
    return true;
  else
    return false;
}

public void repositionFrontEdge() {
  SubCube front = getFrontTopEdge();
  SubCube back = getBackTopEdge();
  SubCube left = getLeftTopEdge();
  SubCube right = getRightTopEdge();
  while(front.front != cyellow) {
    rubeCube.rotateRowY(-1); 
    front = getFrontTopEdge();
    back = getBackTopEdge();
    left = getLeftTopEdge();
    right = getRightTopEdge();
  }
}

public void solveRedTops() {
  ArrayList<SubCube> cubes = findRedEdges();
  int count = calcRedTops(cubes);
  while(count != 4) {
    cubes = findRedEdges();
    count = calcRedTops(cubes);
    if(count == 0 || count == 1) {
      SubCube front = getFrontTopEdge();
      SubCube right = getRightTopEdge();
      while(front.front != cred 
              && right.right != cred) {
        rubeCube.rotateRowY(-1);
        right = getRightTopEdge();
        front = getFrontTopEdge();
      }
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
    }
    else if(count == 2) {
      boolean ready = false;
      while(!ready) {
        rubeCube.rotateRowY(-1);
        SubCube left = getLeftTopEdge();
        SubCube back = getBackTopEdge();
        SubCube right = getRightTopEdge();
        if(left.top == cred && back.top == cred) {
          ready = true;
        }
        else if(left.top == cred 
                  && right.top == cred) {
          ready = true;
        }
      }
      rubeCube.rotateRowZ(1);
      rubeCube.rotateRowX(1);
      rubeCube.rotateReverseRowY(-1);
      rubeCube.rotateReverseRowX(1);
      rubeCube.rotateRowY(-1);
      rubeCube.rotateReverseRowZ(1);
    }
  }
}

public SubCube getLeftTopEdge() {
  for(SubCube sc: rubeCube.cubes) {
    PVector pos = sc.currentPos;
    if(pos.equals(new PVector(-1, -1, 0))) 
      return sc;
  } 
  return null;
}

public SubCube getBackTopEdge() {
  for(SubCube sc: rubeCube.cubes) {
    PVector pos = sc.currentPos;
    if(pos.equals(new PVector(0, -1, -1))) 
      return sc;
  } 
  return null;
}

public SubCube getFrontTopEdge() {
  for(SubCube sc: rubeCube.cubes) {
    PVector pos = sc.currentPos;
    if(pos.equals(new PVector(0, -1, 1))) 
      return sc;
  } 
  return null;
}

public SubCube getRightTopEdge() {
  for(SubCube sc: rubeCube.cubes) {
    PVector pos = sc.currentPos;
    if(pos.equals(new PVector(1, -1, 0))) 
      return sc;
  } 
  return null;
}

public SubCube findFrontTopEdge() {
  int red = 0;
  int yellow = 0;
  int white = 0;
  for(SubCube sc: rubeCube.cubes) {
    red = 0;
    yellow = 0;
    white = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cred)
        red += 1;
      else if(colors[i] == cwhite)
        white += 1; 
      else if(colors[i] == cyellow)
        yellow += 1; 
    }
    if(red == 1 && white == 4 && yellow == 1)
      return sc;
  } 
  return null;
}

public int calcRedTops(ArrayList<SubCube> list) {
  int count = 0;
  for(int i = 0; i < list.size(); i++) {
    if(list.get(i).top == cred) {
      count++; 
    }
  } 
  return count;
}

public ArrayList<SubCube> findRedEdges() {
  int red = 0;
  int white = 0;
  int someColor = 0;
  ArrayList<SubCube> list = new ArrayList<SubCube>();
  for(SubCube sc: rubeCube.cubes) {
    red = 0;
    white = 0;
    someColor = 0;
    int[] colors = sc.makeListOfFaces();
    for(int i = 0; i < colors.length; i++) {
      if(colors[i] == cwhite)
        white += 1;
      else if(colors[i] == cred)
        red += 1; 
      else if(colors[i] != cred 
                && colors[i] != cwhite)
        someColor += 1; 
    }
    if(white == 4 && red == 1 && someColor == 1)
      list.add(sc);
  } 
  return list;
}

public void checkerBoardPattern() {
  if(!rubeCube.winCheckSum())
    solve();
  Cube oldcube = rubeCube;
  rubeCube = rubeCube.solveableClone();
  rubeCube.rotateRowY(-1);
  rubeCube.rotateRowY(-1);
  rubeCube.rotateRowY(1);
  rubeCube.rotateRowY(1);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowZ(0);
  rubeCube.rotateRowZ(0);
  rubeCube = oldcube;
}

public void checkerFlowerPattern() {
  if(!rubeCube.winCheckSum())
    solve();
  Cube oldcube = rubeCube;
  rubeCube = rubeCube.solveableClone();
  rubeCube.rotateRowY(-1);
  rubeCube.rotateRowY(-1);
  rubeCube.rotateRowY(1);
  rubeCube.rotateRowY(1);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowZ(0);
  rubeCube.rotateRowZ(0);
  rubeCube.rotateRowY(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateReverseRowY(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateRowX(0);
  rubeCube = oldcube;
}

public void middleSwitchPattern() {
  if(!rubeCube.winCheckSum())
    solve();
  Cube oldcube = rubeCube;
  rubeCube = rubeCube.solveableClone();
  rubeCube.rotateRowY(0);
  rubeCube.rotateRowX(0);
  rubeCube.rotateReverseRowY(0);
  rubeCube.rotateReverseRowX(0);
  rubeCube = oldcube;
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "CUBE" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
